/*
 *  expr/transcript.c
 *
 *  $Author: baptiste $, $Date: 2008-05-13 15:33:52 $, $Version$
 *
 *  Libgdl : a C library for statistical genetics
 *
 *  Copyright (C) 2009  Jean-Baptiste Veyrieras, Univeristy of Chicago
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
 */
#include <gdl/gdl_common.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_string.h>
#include <gdl/gdl_sort.h>
#include <gdl/gdl_statistics.h>
#include <gdl/gdl_math.h>
#include <gdl/gdl_rng.h>
#include <gdl/gdl_randist.h>
#include <gdl/gdl_specfunc.h>
#include <gdl/gdl_sort_double.h>
#include <gdl/gdl_statistics_double.h>
#include <gdl/gdl_expr_transcript.h>

gdl_expr_transcript *
gdl_expr_transcript_alloc (const gdl_string * name, const size_t nexon)
{
	gdl_expr_transcript * p;

	p = GDL_CALLOC (gdl_expr_transcript, 1);

	p->name = gdl_string_clone (name);

	if (nexon)
	{
		p->nexon    = nexon;
		p->exon_idx = GDL_CALLOC (size_t, nexon);
	}

	return p;
}

void
gdl_expr_transcript_free (gdl_expr_transcript * p)
{
	if (p)
	{
		gdl_string_free (p->name);
		GDL_FREE (p->exon_idx);
		GDL_FREE (p->probe_idx);
		GDL_FREE (p);
	}
}

size_t
gdl_expr_transcript_add_probe (gdl_expr_transcript * p, const size_t e)
{
	if (p->nprobe)
	{
		size_t * tmp = GDL_MALLOC (size_t, p->nprobe + 1);
		memcpy(tmp, p->probe_idx, sizeof(size_t)*p->nprobe);
		GDL_FREE (p->probe_idx);
		p->probe_idx = tmp;
	}
	else
	{
		p->probe_idx = GDL_MALLOC (size_t, 1);
	}
	p->probe_idx[p->nprobe] = e;

	return ++(p->nprobe);
}


gdl_expr_transcript *
gdl_expr_transcript_fread (FILE * stream)
{
	if (stream)
	{
		int status;
		gdl_expr_transcript * p;

		p = GDL_CALLOC (gdl_expr_transcript, 1);

		p->name = gdl_string_fread (stream);
		GDL_FREAD_STATUS (p->name != 0, 1);
		status = fread(&(p->nexon), sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread(&(p->nprobe), sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);

		if (p->nexon)
		{
			p->exon_idx = GDL_MALLOC (size_t, p->nexon);
			status = fread(p->exon_idx, sizeof(size_t), p->nexon, stream);
			GDL_FREAD_STATUS (status, p->nexon);
		}
		if (p->nprobe)
		{
			p->probe_idx = GDL_MALLOC (size_t, p->nprobe);
			status = fread(p->probe_idx, sizeof(size_t), p->nprobe, stream);
			GDL_FREAD_STATUS (status, p->nprobe);
		}

		status = fread(&(p->txStart), sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread(&(p->txEnd), sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread(&(p->cdsStart), sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fread(&(p->cdsEnd), sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);

		return p;
	}
	return 0;
}

int
gdl_expr_transcript_fwrite (FILE * stream, const gdl_expr_transcript * p)
{
	if (stream && p)
	{
		int status;

		status = gdl_string_fwrite (stream, p->name);
		GDL_FREAD_STATUS (status, GDL_SUCCESS);
		status = fwrite(&(p->nexon), sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);
		status = fwrite(&(p->nprobe), sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);

		if (p->nexon)
		{
			status = fwrite(p->exon_idx, sizeof(size_t), p->nexon, stream);
			GDL_FWRITE_STATUS (status, p->nexon);
		}
		if (p->nprobe)
		{
			status = fwrite(p->probe_idx, sizeof(size_t), p->nprobe, stream);
			GDL_FWRITE_STATUS (status, p->nprobe);
		}

		status = fwrite(&(p->txStart), sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fwrite(&(p->txEnd), sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fwrite(&(p->cdsStart), sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);
		status = fwrite(&(p->cdsEnd), sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);

		return GDL_SUCCESS;
	}
	return GDL_EINVAL;
}
