/*
 *  igdb/transreg.c 
 * 
 *  $Author: baptiste $, $Date: 2008-05-13 15:33:47 $, $Version$
 *  
 *  Libgdl : a C library for statistical genetics
 * 
 *  Copyright (C) 2009 Jean-Baptiste Veyrieras, University of Chicago, USA.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */
#include <math.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_string.h>
#include <gdl/gdl_list.h>
#include <gdl/gdl_matrix.h>
#include <gdl/gdl_labeled_matrix.h>
#include <gdl/gdl_vector.h>
#include <gdl/gdl_blas.h>
#include <gdl/gdl_cblas.h>
#include <gdl/gdl_math.h>
#include <gdl/gdl_rng.h>
#include <gdl/gdl_randist.h>
#include <gdl/gdl_specfunc.h>
#include <gdl/gdl_sort_double.h>
#include <gdl/gdl_statistics_double.h>
#include <gdl/gdl_bayesian_regression.h>
#include <gdl/gdl_snp_annotation.h>
#include <gdl/gdl_igdb_chromosome.h>
#include <gdl/gdl_expr_feature_cis.h>
#include <gdl/gdl_igdb_snp_reg.h>

gdl_igdb_reg_snp_record * 
gdl_igdb_reg_snp_record_alloc (const gdl_string * chrom_name, const gdl_string * snp_name, const long snp_pos, const double snp_maf, const size_t N, const double a, const double r2, const double pval, const double bf)
{
	gdl_igdb_reg_snp_record * r;
	
	r = GDL_MALLOC (gdl_igdb_reg_snp_record, 1);
	
	r->chrom_name = gdl_string_clone (chrom_name);
	r->snp_name = gdl_string_clone (snp_name);
	r->snp_pos = snp_pos;
	r->snp_maf = snp_maf;
	r->N = N;
	r->a = a;
 	r->r2 = r2;
 	r->pval = pval;
 	r->bf = bf;
 	
 	return r;
}

gdl_igdb_reg_snp_record * 
gdl_igdb_reg_snp_record_alloc2 (void)
{
	gdl_igdb_reg_snp_record * r;
	
	r = GDL_CALLOC (gdl_igdb_reg_snp_record, 1);
	
 	return r;
}

void 
gdl_igdb_reg_snp_record_update (gdl_igdb_reg_snp_record * r, const gdl_string * chrom_name, const gdl_string * snp_name, const long snp_pos, double snp_maf, const size_t N, const double a, const double r2, const double pval, const double bf)
{
	gdl_string_free (r->chrom_name);
	gdl_string_free (r->snp_name);
	r->chrom_name = gdl_string_clone (chrom_name);
	r->snp_name = gdl_string_clone (snp_name);
	r->snp_pos = snp_pos;
	r->snp_maf = snp_maf;
	r->N = N;
	r->a = a;
 	r->r2 = r2;
 	r->pval = pval;
 	r->bf = bf;
}

void
gdl_igdb_reg_snp_record_free (gdl_igdb_reg_snp_record * r)
{
	if (r)
	{
		gdl_string_free (r->chrom_name);
		gdl_string_free (r->snp_name);
		GDL_FREE (r);	
	}
}

int 
gdl_igdb_reg_snp_record_fprintf (FILE * stream, const gdl_igdb_reg_snp_record * r)
{
	fprintf (stream, "%s %s %d %1.4f %d %g %g %g %g", r->chrom_name, r->snp_name, r->snp_pos, r->snp_maf, r->N, r->a, r->r2, r->pval, r->bf);
}

gdl_igdb_reg_best_buffer * 
gdl_igdb_reg_best_buffer_alloc (const gdl_labeled_matrix * Y,
                                const int y_from,
                                const int y_to,
                                const gdl_feature_table * feature_table)
{
	size_t i;
	gdl_igdb_reg_best_buffer * b;
	
	b = GDL_CALLOC (gdl_igdb_reg_best_buffer, 1);
	
	b->Y = Y;
	b->feature_table = feature_table;
	
	b->from = (y_from == -1) ? 0 : y_from;
	b->to   = (y_to == -1) ? Y->X->size1 : y_to;
	b->size = b->to - b->from;
	if (b->size)
	{
		b->nbf = GDL_CALLOC (size_t, b->size);
		b->npv = GDL_CALLOC (size_t, b->size);
		b->bf_records = GDL_CALLOC (gdl_igdb_reg_snp_record **, b->size);
		b->pv_records = GDL_CALLOC (gdl_igdb_reg_snp_record **, b->size);
		b->bf_max  = GDL_CALLOC (double, b->size);
		b->pv_min  = GDL_CALLOC (double, b->size);
	}
	return b;
}

void 
gdl_igdb_reg_best_buffer_free (gdl_igdb_reg_best_buffer * b)
{
	if (b)
	{
		if (b->size)
		{
			size_t i, j;
			GDL_FREE (b->bf_max);
			GDL_FREE (b->pv_min);
			for(i = 0; i < b->size; i++)
			{
				if (b->bf_records[i])
				{
					for(j = 0; j < b->nbf[i]; j++)
					{
						gdl_igdb_reg_snp_record_free (b->bf_records[i][j]);
					}
					GDL_FREE (b->bf_records[i]);
				}
				if (b->pv_records[i])
				{
					for(j = 0; j < b->npv[i]; j++)
					{
						gdl_igdb_reg_snp_record_free (b->pv_records[i][j]);
					}
					GDL_FREE (b->pv_records[i]);
				}
			}
			GDL_FREE (b->bf_records);
			GDL_FREE (b->pv_records);
			GDL_FREE (b->nbf);
			GDL_FREE (b->npv);			
		}
		GDL_FREE (b);
	}
}

// return the snp record to fill if of
// otherwise NULL
gdl_igdb_reg_snp_record * 
gdl_igdb_reg_best_buffer_test_bf (gdl_igdb_reg_best_buffer * b, const size_t fx, const double bf, const unsigned char mode)
{
	if (b->bf_max[fx] < bf)
	{
		size_t i;
		switch(mode)
		{
			case 'p':
				b->bf_max[fx] = bf;
				b->nbf[fx]    = 1;
				return 0;
			default:
				for(i = 0; i < b->nbf[fx]; i++)
				{
					gdl_igdb_reg_snp_record_free (b->bf_records[fx][i]);
				}
				GDL_FREE (b->bf_records[fx]);
				b->bf_max[fx] = bf;
				b->nbf[fx]    = 1;
				b->bf_records[fx]    = GDL_MALLOC (gdl_igdb_reg_snp_record *, 1);
				b->bf_records[fx][0] = gdl_igdb_reg_snp_record_alloc2();
			return b->bf_records[fx][0];
		}
	}
	else if (b->bf_max[fx] == bf)
	{
		gdl_igdb_reg_snp_record ** tmp;
		switch(mode)
		{
			case 'p':
				(b->nbf[fx])++;
				return 0;
			default:
				tmp = GDL_MALLOC (gdl_igdb_reg_snp_record *, b->nbf[fx]+1);
				memcpy(tmp, b->bf_records[fx], sizeof(gdl_igdb_reg_snp_record *)*b->nbf[fx]);
				GDL_FREE (b->bf_records[fx]);
				b->bf_records[fx] = tmp;
				b->bf_records[fx][b->nbf[fx]] = gdl_igdb_reg_snp_record_alloc2();
				(b->nbf[fx])++;
				return b->bf_records[fx][b->nbf[fx]];
		}
	}
	return 0;
}

gdl_igdb_reg_snp_record * 
gdl_igdb_reg_best_buffer_test_pv (gdl_igdb_reg_best_buffer * b, const size_t fx, const double pv, const unsigned char mode)
{
	if (b->pv_min[fx] > pv)
	{
		size_t i;
		switch(mode)
		{
			case 'p':
				b->pv_min[fx] = pv;
				b->npv[fx]    = 1;
				return 0;
			default:
				for(i = 0; i < b->npv[fx]; i++)
				{
					gdl_igdb_reg_snp_record_free (b->pv_records[fx][i]);
				}
				GDL_FREE (b->pv_records[fx]);
				b->pv_min[fx] = pv;
				b->npv[fx]    = 1;
				b->pv_records[fx]    = GDL_MALLOC (gdl_igdb_reg_snp_record *, 1);
				b->pv_records[fx][0] = gdl_igdb_reg_snp_record_alloc2();
				return b->pv_records[fx][0];
		}
	}
	else if (b->pv_min[fx] == pv)
	{
		gdl_igdb_reg_snp_record ** tmp;
		switch(mode)
		{
			case 'p':
				(b->npv[fx])++;
				return 0;
			default:
				tmp = GDL_MALLOC (gdl_igdb_reg_snp_record *, b->npv[fx]+1);
				memcpy(tmp, b->pv_records[fx], sizeof(gdl_igdb_reg_snp_record *)*b->npv[fx]);
				GDL_FREE (b->pv_records[fx]);
				b->pv_records[fx] = tmp;
				b->pv_records[fx][b->npv[fx]] = gdl_igdb_reg_snp_record_alloc2();
				(b->npv[fx])++;
				return b->bf_records[fx][b->npv[fx]];
		}
	}
	return 0;
}

int 
gdl_igdb_reg_best_buffer_fprintf (FILE * stream, const gdl_igdb_reg_best_buffer * b, const unsigned char mode)
{
	size_t i, j, k;
	gdl_igdb_reg_snp_record * rec;
	
	for(j = 0, i = b->from; i < b->to; i++)
	{
		if (b->nbf[j])
		{
			for(k = 0; k < b->nbf[k]; k++)
			{
				switch(mode)
				{
					case 'p':
						fprintf (stream, "bf %d %g", b->nbf[j], b->bf_max[j]);
						break;
					default:
						rec = b->bf_records[j][k];
						fprintf (stream, "%s ", b->Y->row_names[i]);
						gdl_igdb_reg_snp_record_fprintf (stream, rec);
						fprintf (stream, " bf\n");
						break;
				}
			}
		}
		if (b->npv[j])
		{
			for(k = 0; k < b->npv[k]; k++)
			{
				switch(mode)
				{
					case 'p':
						fprintf (stream, "pv %d %g", b->npv[j], b->pv_min[j]);
						break;
					default:
						rec = b->pv_records[j][k];
						fprintf (stream, "%s ", b->Y->row_names[i]);
						gdl_igdb_reg_snp_record_fprintf (stream, rec);
						fprintf (stream, " pv\n");
						break;
				}
			}
		}
		j++;
	}
}
