/*
 *  igdb/chrom.c 
 * 
 *  $Author: baptiste $, $Date: 2008-05-13 15:33:47 $, $Version$
 *  
 *  Libgdl : a C library for statistical genetics
 * 
 *  Copyright (C) 2009 Jean-Baptiste Veyrieras, University of Chicago, USA.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */

#include <gdl/gdl_common.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_string.h>
#include <gdl/gdl_hash.h>
#include <gdl/gdl_table.h>
#include <gdl/gdl_labeled_matrix.h>
#include <gdl/gdl_snp_genome.h>
#include <gdl/gdl_expr_genome.h>
#include <gdl/gdl_expr_feature.h>
#include <gdl/gdl_expr_feature_cis.h>
#include <gdl/gdl_igdb_chromosome.h>


gdl_igdb_chromosome * 
gdl_igdb_chromosome_alloc (const gdl_string * name, const size_t idx)
{
	gdl_igdb_chromosome * c;
	
	c = GDL_CALLOC (gdl_igdb_chromosome, 1);
	
	c->name = gdl_string_clone (name);
	
	c->idx = idx;
	
	return c;
}

void 
gdl_igdb_chromosome_free (gdl_igdb_chromosome * c)
{
	if (c)
	{
		gdl_string_free (c->name);
		if (c->snp_data)
		{
			gdl_snp_chromosome_free (c->snp_data);
		}
		if (c->ftr_data)
		{
			gdl_expr_feature_chromosome_free (c->ftr_data, c->exp_data);
		}
		if (c->ftr_cis)
		{
			gdl_expr_feature_cis_chromosome_free (c->ftr_cis);
		}
		if (c->exp_data)
		{
			gdl_expr_chromosome_free (c->exp_data);
		}
		GDL_FREE (c);
	}	
}

void
gdl_igdb_chromosome_flag_snp_on_ftr_ends (gdl_igdb_chromosome * chrom)
{
	if (!chrom->ftr_data || !chrom->snp_data)
	{
		return;	
	}
	const size_t nsnp     = chrom->snp_data->chrom->size;
	const size_t nfeature = chrom->ftr_data->nfeature;
	long * starts, * ends;
	size_t i, j, k, l, ns, ne, nbloc=0; 
	gdl_snp ** snps              = chrom->snp_data->chrom->snps;
	gdl_expr_feature ** features = chrom->ftr_data->features;
	
	if (chrom->ftr_snp_start || chrom->ftr_snp_end)
	{
		GDL_FREE (chrom->ftr_snp_start);
		GDL_FREE (chrom->ftr_snp_end);	
	}
	
	chrom->ftr_snp_start = GDL_CALLOC (size_t, nfeature);
	chrom->ftr_snp_end = GDL_CALLOC (size_t, nfeature);
	
	starts = GDL_MALLOC (long, nfeature);
	ends   = GDL_MALLOC (long, nfeature); 
	
	for (k = 0; k < nfeature; k++)
	{
		const long * fs = (features[k]->T->get_starts)(chrom->exp_data, features[k]->idx, &ns);
		const long * fe = (features[k]->T->get_ends)(chrom->exp_data, features[k]->idx, &ne);
		starts[k] = fs[0];
		ends[k]   = fe[ne-1];
	} 
	
	for (j = 0; j < nsnp; j++)
	{ 
		for (k = 0; k < nfeature; k++)
		{
			if (chrom->ftr_snp_end[k])
			{
				continue;	
			}
			if (snps[j]->position < starts[k])
			{
				continue;
			}
			else if (snps[j]->position > ends[k] && !chrom->ftr_snp_end[k])
			{
				chrom->ftr_snp_end[k] = j;
				if (!chrom->ftr_snp_start[k]) 
				{
					chrom->ftr_snp_start[k] = j;
				}
			}
			else if (!chrom->ftr_snp_start[k])
			{
				chrom->ftr_snp_start[k] = j+1;
			}
			//printf ("%d %d %d (%d %d)\n", k, chrom->ftr_snp_start[k], chrom->ftr_snp_end[k], starts[k], ends[k]);
		}
	}
	for (k = 0; k < nfeature; k++)
	{
		if (chrom->ftr_snp_start[k]==0 && chrom->ftr_snp_end[k]==0)
		{
			continue;
		}
		else if (chrom->ftr_snp_start[k]) 
		{
			(chrom->ftr_snp_start[k])--;
			if (chrom->ftr_snp_end[k]) 
			{
				(chrom->ftr_snp_end[k])--;
			}
			else if ( chrom->ftr_snp_start[k]!=0)
			{
				chrom->ftr_snp_end[k] = nsnp-1;
			}
		}
		else if (chrom->ftr_snp_end[k])
		{
			(chrom->ftr_snp_end[k])--;
			chrom->ftr_snp_start[k] = chrom->ftr_snp_end[k];
		}
	}
	
	GDL_FREE (starts);
	GDL_FREE (ends);
}
	
void
gdl_igdb_chromosome_flag_snp_on_ftr_window_end (gdl_igdb_chromosome * chrom, const long window_size)
{
	if (!chrom->ftr_data || !chrom->snp_data)
	{
		return;	
	}
	const size_t nsnp     = chrom->snp_data->chrom->size;
	const size_t nfeature = chrom->ftr_data->nfeature;
	int k;
	size_t i, j, l, ne, ns;
	gdl_snp ** snps              = chrom->snp_data->chrom->snps;
	gdl_expr_feature ** features = chrom->ftr_data->features;
	
	if (chrom->ftr_snp_up || chrom->ftr_snp_down)
	{
		GDL_FREE (chrom->ftr_snp_up);
		GDL_FREE (chrom->ftr_snp_down);	
	}
	
	chrom->ftr_snp_up = GDL_CALLOC (size_t, nfeature);
	chrom->ftr_snp_down = GDL_CALLOC (size_t, nfeature); 
	
	for (i = 0; i < nfeature; i++)
	{
		if (chrom->ftr_snp_start[i] && chrom->ftr_snp_end[i])
		{
			
			const long * fs = (features[i]->T->get_starts)(chrom->exp_data, features[i]->idx, &ns);
			const long * fe = (features[i]->T->get_ends)(chrom->exp_data, features[i]->idx, &ne);
			
			for (k = chrom->ftr_snp_start[i]; k > 0 && snps[k-1]->position >= (fs[0] - window_size);k--);
			//printf ("k=%d UP::%d\n", k, snps[k]->position - fs[0]);
			chrom->ftr_snp_up[i] = k;
			//if (block->snp_up == block->snp_start) block->snp_up++;
			//printf (">> %d %d (%d)\n", block->end, snps[block->snp_end]->position, block->snp_end);
			for (k = chrom->ftr_snp_end[i] + 1; k < nsnp && snps[k]->position <= fe[ne-1] + window_size;k++);
			chrom->ftr_snp_down[i] = k-1;
			//if (block->snp_down == block->snp_end) block->snp_down--;
			//printf ("k=%d DOWN::%d\n", k, snps[k-1]->position - fe[ne-1]);
			//if (block->snp_up==0 && block->snp_down==nsnp-1)
			//	block->snp_up=block->snp_down=nsnp-1;
		}
	}	
}

FILE *
gdl_igdb_chromosome_set_logger (gdl_igdb_chromosome * g, FILE * stream)
{
	FILE * old = g->logger;
	g->logger = stream;
	return old;
}
