/*
 *  igdb/cisreg.c 
 * 
 *  $Author: baptiste $, $Date: 2008-05-13 15:33:47 $, $Version$
 *  
 *  Libgdl : a C library for statistical genetics
 * 
 *  Copyright (C) 2009 Jean-Baptiste Veyrieras, University of Chicago, USA.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */
#include <math.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_string.h>
#include <gdl/gdl_matrix.h>
#include <gdl/gdl_vector.h>
#include <gdl/gdl_blas.h>
#include <gdl/gdl_cblas.h>
#include <gdl/gdl_math.h>
#include <gdl/gdl_rng.h>
#include <gdl/gdl_randist.h>
#include <gdl/gdl_specfunc.h>
#include <gdl/gdl_sort_double.h>
#include <gdl/gdl_statistics_double.h>
#include <gdl/gdl_bayesian_regression.h>
#include <gdl/gdl_snp_annotation.h>
#include <gdl/gdl_igdb_chromosome.h>
#include <gdl/gdl_expr_feature_cis.h>
#include <gdl/gdl_igdb_snp_reg.h>

#define VECTOR_SET(X,i,y)(*(X->data+i*X->stride)=y)
#define VECTOR_GET(X,i)(*(X->data+i*X->stride))
#define MATRIX_SET(X,i,j,y)(*(X->data+(i*X->tda+j))=y)
#define MATRIX_GET(X,i,j)(*(X->data+(i*X->tda+j)))

size_t
gdl_igdb_cis_reg_snp_init (gdl_bayreg_workspace * bayes, gdl_igdb_chromosome * chrom, const size_t snp, double popf[], double * f, size_t * rm_individual[])
{
	const size_t P = chrom->ploidy;
	size_t N = chrom->nindiv;
	size_t i, j, k, kk, l, p, npop;
	size_t nmis = 0, nrm = 0;
	double a, d;
	
	// first, how many individuals ?
	if (rm_individual)
	{
		for(i = 0; i < chrom->npop; i++)
		{
			for(j = 0; j < chrom->pop_sizes[i]; j++)
			{
				N -= rm_individual[i][j];
			}	
		}
	}
	bayes->y->size = bayes->X->size1 = N;
	// then, how many populations ?
	for (npop = i = 0; i < chrom->npop; i++)
	{
		if (gdl_snp_chromosome_is_polymorphic (chrom->snp_data, i, snp))
		{
			if (rm_individual)
			{
				for(j = 0; j < chrom->pop_sizes[i]; j++)
				{
					if (!rm_individual[i][j])
						break;
				}
				if (j < chrom->pop_sizes[i])
				{
					npop++;
				}
			}
			else
			{
				npop++;
			}
		}
	}
	if (npop)
	{
		bayes->P = npop;
		*f = 0;
		for (kk = k = i = 0; i < chrom->npop; i++)
		{
			popf[i] = 0;
			if (gdl_snp_chromosome_is_polymorphic (chrom->snp_data, i, snp))
			{
				for(j = 0; j < chrom->pop_sizes[i]; j++, kk++)
				{
					if (rm_individual && rm_individual[i][j])
					{
						nrm++;
						continue;
					}
					if (gdl_snp_chromosome_is_missing (chrom->snp_data, i, j, snp))
					{
						nmis++;
						continue;
					}
					if (gdl_isnan (gdl_matrix_get (bayes->Y, kk, 0)))
					{
						nmis++;
						continue;
					}
					
					VECTOR_SET (bayes->y, k, gdl_matrix_get (bayes->Y, kk, 0));
					
					gdl_snp_chromosome_get_adddom (chrom->snp_data, i, j, snp, &a, &d);
					
					MATRIX_SET(bayes->X, k, 1, a);
					MATRIX_SET(bayes->X, k, 2, d);
					
					//printf ("%d %g %g %d\n", snp, VECTOR_GET (bayes->y, k), MATRIX_GET (bayes->X, k, 1), gdl_snp_chromosome_get_genotype(chrom->snp_data, i, j, snp));
					
					popf[i] += a / P;
					
					k++;
					
				}
				if (nmis+nrm == chrom->pop_sizes[i])
				{
					*f = 0;
				}
				else
				{ 
					*f += popf[i];
					popf[i] /= (chrom->pop_sizes[i]-nmis-nrm);
				}
			}
			else kk += chrom->pop_sizes[i];
		}
		bayes->y->size = bayes->X->size1 = k;
		*f /= k;
	}
	
	//printf ("k = %d\n", k);
	//fflush (stdout);
	
	return (*f > 0.0 && *f < 1.0) ? npop : 0;
}

// REMOVE PIDX
void
gdl_igdb_cis_reg_feature_init (gdl_bayreg_workspace * bayes,
                               gdl_igdb_chromosome * chrom,
                               gdl_expr_feature * feature,
                               const gdl_boolean qnorm)
{
	size_t i, j, k, p, * rk = 0;
	double * x = 0;

	bayes->np = 1;
		
	gdl_matrix_free (bayes->Y);
	
	bayes->Y = gdl_matrix_alloc (chrom->nindiv, 1);
	
	//printf ("NINDIV = %d\n", chrom->nindiv);
	
	for (k = i = 0; i < chrom->npop; i++)
	{
		const size_t n = chrom->pop_sizes[i];
		if (qnorm)
		{
			double * x = gdl_stats_qqnorm (feature->data[i], n);
			for(j = 0; j < n; j++, k++)
			{
				MATRIX_SET (bayes->Y, k, 0, x[j]);
			}
			GDL_FREE (x);
		}
		else
		{
			for(j = 0; j < n; j++, k++)
			{
				MATRIX_SET (bayes->Y, k, 0, feature->data[i][j]);
			}
		}
	}
}

gdl_expr_feature_cis_chromosome * 
gdl_igdb_chromosome_cis_reg_run (gdl_igdb_chromosome * chrom,
                              gdl_bayreg_model * model,
                              const gdl_boolean qnorm,
                              size_t * rm_individual[],
                              const gdl_feature_table * gene_set)
{
	const size_t G   = chrom->ftr_data->nfeature;
	const size_t P   = chrom->ploidy;
	const size_t N   = chrom->nindiv;
	const size_t POP = chrom->npop;
	size_t i, j, k, l, g, m, n, maxg, npop, msize;
	double max, bf, u, a, d, * popf, f, r2, pval;
	gdl_expr_feature * feature;
	gdl_snp ** snps;
	gdl_bayreg_workspace * bayes;
	gdl_expr_feature_cis_chromosome * rchrom;
	gdl_expr_feature_cis            * rfeature;
	gdl_expr_feature_cis_snp        * rsnp;
	
	snps = chrom->snp_data->chrom->snps;
	
	msize = gdl_bayreg_model_bf_storage_size (model);
	
	//printf ("N = %d POP = %d\n", N, POP);fflush(stdout);
	
	bayes = gdl_bayreg_workspace_alloc (N, POP);
	
	popf = GDL_CALLOC (double, POP);
	
	if (chrom->logger)
	{
		fprintf (chrom->logger, ">%s |nfeature=%d|nsnp=%d|npop=%d|nindiv=%d|\n", chrom->name, G, chrom->snp_data->chrom->size, POP, N);
		fflush(chrom->logger);
	}
	
	rchrom = gdl_expr_feature_cis_chromosome_alloc (G);
	
	for (i = 0; i < G; i++)
	{
		feature = chrom->ftr_data->features[i];
		
		if (gene_set && !gdl_feature_table_is_inside (gene_set, chrom->name, feature->name))
		{
			rchrom->features[i] = gdl_expr_feature_cis_alloc (-1, -1);
			continue;
		}
		else
		{
			rfeature = gdl_expr_feature_cis_alloc (chrom->ftr_snp_up[i], chrom->ftr_snp_down[i]);	
		}
		
		gdl_igdb_cis_reg_feature_init (bayes, chrom, feature, qnorm);
			
		for (l = chrom->ftr_snp_up[i]; l <= chrom->ftr_snp_down[i]; l++)
		{
			rsnp   = 0;
			
			npop   = gdl_igdb_cis_reg_snp_init (bayes, chrom, l, popf, &f, rm_individual);
			
			//printf ("%d %d %d N = %d ==> %p\n", l, chrom->ftr_snp_up[i], chrom->ftr_snp_down[i], bayes->y->size, bayes->X);fflush(stdout);
			
			if (npop)
			{
				rsnp = gdl_expr_feature_cis_snp_alloc (msize);
				gdl_bayreg_model_single_bf (model, bayes, rsnp->bf);
				gdl_igdb_reg_perform_snp_add (bayes, &a, &r2, &pval);
				rsnp->pval = pval;
				//printf ("%s %s %g\n", feature->name, snps[l]->rs, pval);
			}
			
			rfeature->snps[ l - chrom->ftr_snp_up[i] ] = rsnp;
			
		} // snp
		
		rchrom->features[i] = rfeature;
		
		if (chrom->logger)
		{
			gdl_progress_bar (chrom->logger, 25, i, G);
		}
		
	} // gene
	
	GDL_FREE (popf);
	
	gdl_bayreg_workspace_free (bayes);
	
	return rchrom;
}

int
gdl_igdb_chromosome_cis_bayes_permut (gdl_igdb_chromosome * chrom,
                                      gdl_bayreg_model * model,
                                      const gdl_boolean qnorm,
                                      const gdl_rng * rng,
                                      const size_t permut,
                                      const gdl_boolean best_snp,
                                      FILE * stream)
{
	const size_t G = chrom->ftr_data->nfeature;
	const size_t P = chrom->ploidy;
	const size_t N = chrom->nindiv;
	const size_t POP = chrom->npop;
	size_t i, j, k, l, p, g, m, n, b, nmax, maxg, npop, msize;
	double max, * tbf, bf, logbf, gwt, u, a, d, * popf, f, r2, pval;
	gdl_expr_feature * feature;
	gdl_snp ** snps;
	gdl_bayreg_workspace * bayes;
	
	snps = chrom->snp_data->chrom->snps;
	
	msize = gdl_bayreg_model_bf_storage_size (model);
	
	bayes = gdl_bayreg_workspace_alloc (N, POP);
	
	popf = GDL_CALLOC (double, POP);
	
	tbf  = GDL_CALLOC (double, msize);
	
	if (chrom->logger)
	{
		fprintf (chrom->logger, ">%s |nfeature=%d|nsnp=%d|npop=%d|nindiv=%d|\n", chrom->name, G, chrom->snp_data->chrom->size, POP, N);
		fflush(chrom->logger);
	}
	
	for (i = 0; i < G; i++)
	{
		feature = chrom->ftr_data->features[i];
		
		gdl_igdb_cis_reg_feature_init (bayes, chrom, feature, qnorm);
			
		for (b = 0; b < permut; b++)
		{
			gdl_igdb_reg_permut (bayes, chrom, rng, !qnorm);
		
			max  = 0.0;
			nmax = 0;
				
			for (l = chrom->ftr_snp_up[i]; l <= chrom->ftr_snp_down[i]; l++)
			{
				npop = gdl_igdb_cis_reg_snp_init (bayes, chrom, l, popf, &f, 0);
				
				if (npop)
				{
					gdl_bayreg_model_single_bf (model, bayes, tbf);
					
					bf = gdl_bayreg_model_average_bf (model, tbf, 0);
					
					if (best_snp)
					{
						if (bf > max)
						{
							max = bf;
							nmax= 1;
						}
						else if (bf == max)
						{
							nmax++;	
						}
					}
					else
					{
						fprintf (stream, "%e\n", bf);
						fflush(stream);
					}
				}
			} // snp
			
			if (best_snp && nmax)
			{
				fprintf (stream, "%d %e\n", nmax, max);
				fflush(stream);	
			}
			
		} // permut
		
		if (chrom->logger)
		{
			gdl_progress_bar (chrom->logger, 25, i, G);
		}
	} // gene
	
	GDL_FREE (popf);
	GDL_FREE (tbf);
	
	gdl_bayreg_workspace_free (bayes);
}

int
gdl_igdb_chromosome_cis_freq_permut (gdl_igdb_chromosome * chrom,
                                     const gdl_boolean qnorm,
                                     const gdl_rng * rng,
                                     const size_t permut,
                                     const gdl_boolean best_snp,
                                     FILE * stream)
{
	const size_t G = chrom->ftr_data->nfeature;
	const size_t P = chrom->ploidy;
	const size_t N = chrom->nindiv;
	const size_t POP = chrom->npop;
	size_t i, j, k, l, p, g, m, n, b, nmax, maxg, npop, msize;
	double max, * tbf, bf, logbf, gwt, u, a, d, * popf, f, r2, pval;
	gdl_expr_feature * feature;
	gdl_snp ** snps;
	gdl_bayreg_workspace * bayes;
	
	snps = chrom->snp_data->chrom->snps;
	
	bayes = gdl_bayreg_workspace_alloc (N, POP);
	
	popf = GDL_CALLOC (double, POP);
	
	if (chrom->logger)
	{
		fprintf (chrom->logger, ">%s |nfeature=%d|nsnp=%d|npop=%d|nindiv=%d|\n", chrom->name, G, chrom->snp_data->chrom->size, POP, N);
		fflush(chrom->logger);
	}
	
	for (i = 0; i < G; i++)
	{
		feature = chrom->ftr_data->features[i];
		
		gdl_igdb_cis_reg_feature_init (bayes, chrom, feature, qnorm);
			
		for (b = 0; b < permut; b++)
		{
			gdl_igdb_reg_permut (bayes, chrom, rng, !qnorm);	
				
			max  = 1.0;
			nmax = 0;
			
			for (l = chrom->ftr_snp_up[i]; l <= chrom->ftr_snp_down[i]; l++)
			{
				npop = gdl_igdb_cis_reg_snp_init (bayes, chrom, l, popf, &f, 0);
				
				if (npop)
				{
					gdl_igdb_reg_perform_snp_add (bayes, &a, &r2, &pval);
					
					if (best_snp)
					{
						if (pval < max)
						{
							max = pval;
							nmax= 1;
						}
						else if (pval == max)
						{
							nmax++;	
						}
					}
					else
					{
						fprintf (stream, "%e\n", pval);
						fflush(stream);
					}
					
				}
			} // snp
			if (best_snp && nmax)
			{
				fprintf (stream, "%d %e\n", nmax, max);
				fflush(stream);	
			}
		}
		
		if (chrom->logger)
		{
			gdl_progress_bar (chrom->logger, 25, i, G);
		}
		
	} // gene
	
	GDL_FREE (popf);
	
	gdl_bayreg_workspace_free (bayes);
}

int
gdl_igdb_cis_reg_residuals (gdl_bayreg_model   * model,
                          gdl_bayreg_workspace *  work,
                          gdl_igdb_chromosome  * chrom,
                          gdl_expr_feature     * feature,
                          const size_t sx)
{
	size_t i, j, k, kk;
	double f, * popf;
	gdl_vector * e;
	
	popf = GDL_MALLOC (double, chrom->npop);
	
	gdl_igdb_cis_reg_snp_init (work, chrom, sx, popf, &f, 0);
	
	e = gdl_bayreg_model_single_residual (model, work);
	
	for (k = kk = i = 0; i < chrom->npop; i++)
	{
		if (popf[i] == 0.0)
		{
			k += chrom->pop_sizes[i];
			continue;
		}
		for(j = 0; j < chrom->pop_sizes[i]; j++, k++, kk++)
		{
			gdl_matrix_set (work->Y, k, 0, gdl_vector_get (e, kk));
		}		
	}	
	
	gdl_vector_free (e); 
	
	GDL_FREE (popf);
}

int 
gdl_igdb_chromosome_cis_reg_partial_run (gdl_igdb_chromosome * chrom, gdl_bayreg_model * model, gdl_expr_feature_cis_chromosome * rchrom, const size_t from, const size_t to, const gdl_boolean qnorm, size_t * rm_individual[], const gdl_feature_table * feature_set)
{
	const size_t G   = chrom->ftr_data->nfeature;
	const size_t P   = chrom->ploidy;
	const size_t N   = chrom->nindiv;
	const size_t POP = chrom->npop;
	size_t i, j, k, l, g, m, n, maxg, npop, msize;
	double max, bf, u, a, d, * popf, f, r2, pval;
	gdl_expr_feature * feature;
	gdl_snp ** snps;
	gdl_bayreg_workspace * bayes;
	gdl_expr_feature_cis            * rfeature;
	gdl_expr_feature_cis_snp        * rsnp;
	
	snps = chrom->snp_data->chrom->snps;
	
	msize = model->size;
	
	bayes = gdl_bayreg_workspace_alloc (N, POP);
	
	popf = GDL_CALLOC (double, POP);
	
	if (chrom->logger)
	{
		fprintf (chrom->logger, ">%s |feature_from=%d|feature_to=%d|nsnp=%d|npop=%d|nindiv=%d|\n", chrom->name, from+1, to, chrom->snp_data->chrom->size, POP, N);
		fflush(chrom->logger);
	}
	
	for (i = from; i < to; i++)
	{
		feature = chrom->ftr_data->features[i];
		
		if (feature_set && !gdl_feature_table_is_inside (feature_set, chrom->name, feature->name))
		{
			rchrom->features[i] = gdl_expr_feature_cis_alloc (-1, -1);
			continue;
		}
		else
		{
			rfeature = gdl_expr_feature_cis_alloc (chrom->ftr_snp_up[i], chrom->ftr_snp_down[i]);	
		}
		
		gdl_igdb_cis_reg_feature_init (bayes, chrom, feature, qnorm);
			
		for (l = chrom->ftr_snp_up[i]; l <= chrom->ftr_snp_down[i]; l++)
		{
			rsnp   = 0;
			
			npop   = gdl_igdb_cis_reg_snp_init (bayes, chrom, l, popf, &f, rm_individual);
			
			if (npop)
			{
				rsnp = gdl_expr_feature_cis_snp_alloc (msize);
				gdl_bayreg_model_single_bf (model, bayes, rsnp->bf);
				gdl_igdb_reg_perform_snp_add (bayes, &a, &r2, &pval);
				rsnp->pval = pval;
			}
			
			rfeature->snps[ l - chrom->ftr_snp_up[i] ] = rsnp;
			
		} // snp
		
		rchrom->features[i] = rfeature;
		
		if (chrom->logger)
		{
			gdl_progress_bar (chrom->logger, 25, i, G);
		}
		
	} // gene
	
	GDL_FREE (popf);
	
	gdl_bayreg_workspace_free (bayes);
	
	return GDL_SUCCESS;
}

int 
gdl_igdb_chromosome_cis_freq_partial_permut (gdl_igdb_chromosome * chrom, const size_t from, const size_t to, const gdl_boolean qnorm, const gdl_rng * rng, const size_t permut, const gdl_boolean best_snp, FILE * stream)
{
	const size_t G = chrom->ftr_data->nfeature;
	const size_t P = chrom->ploidy;
	const size_t N = chrom->nindiv;
	const size_t POP = chrom->npop;
	size_t i, j, k, l, p, g, m, n, b, nmax, maxg, npop, msize;
	double max, * tbf, bf, logbf, gwt, u, a, d, * popf, f, r2, pval;
	gdl_expr_feature * feature;
	gdl_snp ** snps;
	gdl_bayreg_workspace * bayes;
	
	snps = chrom->snp_data->chrom->snps;
	
	bayes = gdl_bayreg_workspace_alloc (N, POP);
	
	popf = GDL_CALLOC (double, POP);
	
	if (chrom->logger)
	{
		fprintf (chrom->logger, ">%s |feature_from=%d|feature_to=%d|nsnp=%d|npop=%d|nindiv=%d|\n", chrom->name, from+1, to, chrom->snp_data->chrom->size, POP, N);
		fflush(chrom->logger);
	}
	
	for (i = from; i < to; i++)
	{
		feature = chrom->ftr_data->features[i];
		
		gdl_igdb_cis_reg_feature_init (bayes, chrom, feature, qnorm);
			
		for (b = 0; b < permut; b++)
		{
			gdl_igdb_reg_permut (bayes, chrom, rng, !qnorm);	
				
			max  = 1.0;
			nmax = 0;
			
			for (l = chrom->ftr_snp_up[i]; l <= chrom->ftr_snp_down[i]; l++)
			{
				npop = gdl_igdb_cis_reg_snp_init (bayes, chrom, l, popf, &f, 0);
				
				if (npop)
				{
				
					gdl_igdb_reg_perform_snp_add (bayes, &a, &r2, &pval);
					
					if (best_snp)
					{
						if (pval < max)
						{
							max = pval;
							nmax= 1;
						}
						else if (pval == max)
						{
							nmax++;	
						}
					}
					else
					{
						fprintf (stream, "%e\n", pval);
						fflush(stream);
					}
					
				}
			} // snp
			if (best_snp && nmax)
			{
				fprintf (stream, "%d %e\n", nmax, max);
				fflush(stream);	
			}
		}
		
		if (chrom->logger)
		{
			gdl_progress_bar (chrom->logger, 25, i, G);
		}
		
	} // gene
	
	GDL_FREE (popf);
	
	gdl_bayreg_workspace_free (bayes);
}

int 
gdl_igdb_chromosome_cis_bayes_partial_permut (gdl_igdb_chromosome * chrom, gdl_bayreg_model * model, const size_t from, const size_t to, const gdl_boolean qnorm, const gdl_rng * rng, const size_t permut, const gdl_boolean best_snp, FILE * stream)
{
	const size_t G = chrom->ftr_data->nfeature;
	const size_t P = chrom->ploidy;
	const size_t N = chrom->nindiv;
	const size_t POP = chrom->npop;
	size_t i, j, k, l, p, g, m, n, b, nmax, maxg, npop, msize;
	double max, * tbf, bf, logbf, gwt, u, a, d, * popf, f, r2, pval;
	gdl_expr_feature * feature;
	gdl_snp ** snps;
	gdl_bayreg_workspace * bayes;
	
	snps = chrom->snp_data->chrom->snps;
	
	msize = gdl_bayreg_model_bf_storage_size (model);
	
	bayes = gdl_bayreg_workspace_alloc (N, POP);
	
	popf = GDL_CALLOC (double, POP);
	
	tbf  = GDL_CALLOC (double, msize);
	
	if (chrom->logger)
	{
		fprintf (chrom->logger, ">%s |feature_from=%d|feature_to=%d|nsnp=%d|npop=%d|nindiv=%d|\n", chrom->name, from+1, to, chrom->snp_data->chrom->size, POP, N);
		fflush(chrom->logger);
	}
	
	for (i = from; i < to; i++)
	{
		feature = chrom->ftr_data->features[i];
		
		gdl_igdb_cis_reg_feature_init (bayes, chrom, feature, qnorm);
			
		for (b = 0; b < permut; b++)
		{
			gdl_igdb_reg_permut (bayes, chrom, rng, !qnorm);
		
			max  = 0.0;
			nmax = 0;
				
			for (l = chrom->ftr_snp_up[i]; l <= chrom->ftr_snp_down[i]; l++)
			{
				npop = gdl_igdb_cis_reg_snp_init (bayes, chrom, l, popf, &f, 0);
				
				if (npop)
				{
					gdl_bayreg_model_single_bf (model, bayes, tbf);
					
					bf = gdl_bayreg_model_average_bf (model, tbf, 0);
					
					if (best_snp)
					{
						if (bf > max)
						{
							max = bf;
							nmax= 1;
						}
						else if (bf == max)
						{
							nmax++;	
						}
					}
					else
					{
						fprintf (stream, "%e\n", bf);
						fflush(stream);
					}
				}
			} // snp
			
			if (best_snp && nmax)
			{
				fprintf (stream, "%d %e\n", nmax, max);
				fflush(stream);	
			}
			
		} // permut
		
		if (chrom->logger)
		{
			gdl_progress_bar (chrom->logger, 25, i, G);
		}
	} // gene
	
	GDL_FREE (popf);
	GDL_FREE (tbf);
	
	gdl_bayreg_workspace_free (bayes);
}
