/*
 *  igdb/fullreg.c 
 * 
 *  $Author: baptiste $, $Date: 2008-05-13 15:33:47 $, $Version$
 *  
 *  Libgdl : a C library for statistical genetics
 * 
 *  Copyright (C) 2009 Jean-Baptiste Veyrieras, University of Chicago, USA.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA * 
 */
#include <math.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_string.h>
#include <gdl/gdl_list.h>
#include <gdl/gdl_matrix.h>
#include <gdl/gdl_labeled_matrix.h>
#include <gdl/gdl_vector.h>
#include <gdl/gdl_blas.h>
#include <gdl/gdl_cblas.h>
#include <gdl/gdl_math.h>
#include <gdl/gdl_rng.h>
#include <gdl/gdl_randist.h>
#include <gdl/gdl_specfunc.h>
#include <gdl/gdl_sort_double.h>
#include <gdl/gdl_statistics_double.h>
#include <gdl/gdl_bayesian_regression.h>
#include <gdl/gdl_snp_annotation.h>
#include <gdl/gdl_igdb_chromosome.h>
#include <gdl/gdl_expr_feature_cis.h>
#include <gdl/gdl_igdb_snp_reg.h>

#define VECTOR_SET(X,i,y)(*(X->data+i*X->stride)=y)
#define VECTOR_GET(X,i)(*(X->data+i*X->stride))
#define MATRIX_SET(X,i,j,y)(*(X->data+(i*X->tda+j))=y)
#define MATRIX_GET(X,i,j)(*(X->data+(i*X->tda+j)))

size_t
gdl_igdb_reg_snp_init (gdl_bayreg_workspace * bayes,
                       const gdl_snp_chromosome * chrom,
                       const size_t N,
                       const size_t snp,
                       double popf[],
                       double * f,
                       size_t * rm_individual[])
{
	const size_t P = chrom->pops[0]->P;
	size_t NN = N;
	size_t i, j, k, kk, l, p, npop;
	size_t nmis = 0, nrm = 0;
	double a, d;
	
	// first, how many individuals ?
	if (rm_individual)
	{
		for(i = 0; i < chrom->npop; i++)
		{
			for(j = 0; j < chrom->pops[i]->N; j++)
			{
				NN -= rm_individual[i][j];
			}	
		}
	}
	bayes->y->size = bayes->X->size1 = NN;
	// then, how many populations ?
	for (npop = i = 0; i < chrom->npop; i++)
	{
		if (gdl_snp_chromosome_is_polymorphic (chrom, i, snp))
		{
			if (rm_individual)
			{
				for(j = 0; j < chrom->pops[i]->N; j++)
				{
					if (!rm_individual[i][j])
						break;
				}
				if (j < chrom->pops[i]->N)
				{
					npop++;
				}
			}
			else
			{
				npop++;
			}
		}
	}
	if (npop)
	{
		bayes->P = npop;
		*f = 0;
		for (kk = k = i = 0; i < chrom->npop; i++)
		{
			popf[i] = 0;
			if (gdl_snp_chromosome_is_polymorphic (chrom, i, snp))
			{
				for(j = 0; j < chrom->pops[i]->N; j++, kk++)
				{
					if (rm_individual && rm_individual[i][j])
					{
						nrm++;
						continue;
					}
					if (gdl_snp_chromosome_is_missing (chrom, i, j, snp))
					{
						nmis++;
						continue;
					}
					if (gdl_isnan (gdl_matrix_get (bayes->Y, kk, 0)))
					{
						nmis++;
						continue;
					}
					
					VECTOR_SET (bayes->y, k, gdl_matrix_get (bayes->Y, kk, 0));
					
					gdl_snp_chromosome_get_adddom (chrom, i, j, snp, &a, &d);
					
					MATRIX_SET(bayes->X, k, 1, a);
					MATRIX_SET(bayes->X, k, 2, d);
					
					popf[i] += a / P;
					
					k++;
					
				}
				if (nmis+nrm == chrom->pops[i]->N)
				{
					*f = 0;
				}
				else
				{ 
					*f += popf[i];
					popf[i] /= (chrom->pops[i]->N-nmis-nrm);
				}
			}
			else kk += chrom->pops[i]->N;
		}
		bayes->y->size = bayes->X->size1 = k;
		*f /= k;
	}
	
	//printf ("k = %d\n", k);
	//fflush (stdout);
	
	return (*f > 0.0 && *f < 1.0) ? npop : 0;
}

void
gdl_igdb_reg_feature_init (gdl_bayreg_workspace * bayes,
						   const gdl_igdb_chromosome * chrom,
		                   gdl_matrix * Y,
		                   const size_t yi,
		                   const gdl_boolean qnorm)
{
	size_t i, j, k, p, * rk = 0;
	double * x = 0;

	bayes->np = 1;
		
	gdl_matrix_free (bayes->Y);
	
	bayes->Y = gdl_matrix_alloc (chrom->nindiv, 1);
	
	for (k = i = 0; i < chrom->npop; i++)
	{
		const size_t n = chrom->pop_sizes[i];
		if (qnorm)
		{
			double * x = GDL_MALLOC (double, n);
			for(j = 0; j < n; j++)
			{
				x[j] = MATRIX_GET (Y, i, k + j);
			}
			double * xx = gdl_stats_qqnorm (x, n);
			for(j = 0; j < n; j++, k++)
			{
				MATRIX_SET (bayes->Y, k, 0, xx[j]);
			}
			GDL_FREE (xx);
			GDL_FREE (x);
		}
		else
		{
			for(j = 0; j < n; j++, k++)
			{
				MATRIX_SET (bayes->Y, k, 0, MATRIX_GET (Y, i, k));
			}
		}
	}
}

double
gdl_igdb_reg_perform_snp_add (gdl_bayreg_workspace * bayes, double * a, double * r2, double * p)
{
	const size_t N = bayes->X->size1;
	size_t i, j;
	double l0, l1, chi, e, f, g, sig0=0, sig1=0, x=0, y=0, xx=0, xy=0, yy=0;
	
	if (N > 2)
	{
		for (i = 0; i < N; i++)
		{
			e   = MATRIX_GET(bayes->X, i, 1);
			x  += e;
			xx += e*e;
			f   = VECTOR_GET(bayes->y, i);
			y  += f;
			xy += e*f;
		}
		*a=(N*xy-x*y)/(N*xx-x*x);
		y/=N;
		x/=N;
		for (i = 0; i < N; i++)
		{
			e = MATRIX_GET(bayes->X, i, 1);
			f = VECTOR_GET(bayes->y, i);
			sig0 += (f-y)*(f-y);
			g = (f-y)-(*a)*(e-x);
			sig1 += g*g;
		}
		
		chi = (sig0-sig1)*(N - 2)/sig1;
		chi = (chi >= 0) ? chi : 0;
		*p  = gdl_sf_beta_inc (0.5*(N - 2), 0.5,  (N - 2)/((N - 2) + chi));
		*r2 = 1-sig1/sig0;
	}
	else
	{
		*p = 1.0;
		chi = *a = *r2 = 0;	
	}
	
	return chi;
}

/**.
 * Regress the phenotypes Y[1..Y->size1]
 * individuals SNPs[from..to] and output
 * only SNPs fullfilling the criteria  
 */
int
gdl_igdb_chromosome_reg_run (gdl_igdb_chromosome * chrom,
                             gdl_bayreg_model * model,
                             const gdl_boolean qnorm,
                             size_t * rm_individual[],
                             const gdl_feature_table * feature_table,
                             const gdl_labeled_matrix * Y,
                             const int y_from,
                             const int y_to,
                             gdl_igdb_reg_best_buffer * best,
                             const double bf_min,
	                         const double pv_max,
                             FILE * stream)
{
	const size_t G    = chrom->ftr_data->nfeature;
	const size_t P    = chrom->ploidy;
	const size_t N    = chrom->nindiv;
	const size_t POP  = chrom->npop;
	const size_t NSNP = chrom->snp_data->chrom->size;
	int * bfx, * pvx;
	size_t i, j, k, msize, from, to, npop, ok, nbf, npv;
	double * bf, * popf, tbf, f, pval, a, r2, chi, max_bf, min_pv;
	gdl_snp ** snps;
	gdl_bayreg_workspace * bayes;
	
	snps = chrom->snp_data->chrom->snps;
	
	msize = gdl_bayreg_model_bf_storage_size (model);
	
	bayes = gdl_bayreg_workspace_alloc (N, POP);
	
	popf = GDL_CALLOC (double, POP);
	
	bf = GDL_CALLOC (double, msize);
	
	from = (y_from == -1) ? 0 : y_from;
	
	to   = (y_to == -1) ? Y->X->size1 : y_to;
	
	if (chrom->logger)
	{
		fprintf (chrom->logger, ">%s |nfeature=%d|nsnp=%d|npop=%d|nindiv=%d|\n", chrom->name, to-from, NSNP, POP, N);
		fflush(chrom->logger);
	}
	
	for(j = from; j <  to; j++)
	{
		if (feature_table && !gdl_feature_table_is_inside (feature_table, chrom->name, Y->row_names[j]))
		{
			continue;
		}
		
		gdl_igdb_reg_feature_init (bayes, chrom, Y->X, j, qnorm);
			
		for (i = 0; i < NSNP; i++)
		{
			npop = gdl_igdb_reg_snp_init (bayes, chrom->snp_data, N, i, popf, &f, rm_individual);
				
			if (npop)
			{
				gdl_bayreg_model_single_bf (model, bayes, bf);
				gdl_igdb_reg_perform_snp_add (bayes, &a, &r2, &pval);
				tbf = gdl_bayreg_model_average_bf (model, bf, 0);
				if (best)
				{
					gdl_igdb_reg_snp_record * rec;
					
					rec = gdl_igdb_reg_best_buffer_test_bf (best, j, tbf, 'n');
					if (rec)
					{
						gdl_igdb_reg_snp_record_update (rec, chrom->name, snps[i]->rs, snps[i]->position, f, bayes->X->size1, a, r2, pval, tbf);
					}
					rec = gdl_igdb_reg_best_buffer_test_pv (best, j, pval, 'n');
					if (rec)
					{
						gdl_igdb_reg_snp_record_update (rec, chrom->name, snps[i]->rs, snps[i]->position, f, bayes->X->size1, a, r2, pval, tbf);
					}
				}
				if (stream)
				{
					if (tbf > bf_min)
					{
						fprintf (stream, "%s %s %s %d %1.4f %d %g %g %g %g bf\n", Y->row_names[j], chrom->name, snps[i]->rs, snps[i]->position, f, bayes->X->size1, a, r2, pval, tbf);
						fflush (stream);
					}
					if (pval < pv_max)
					{
						fprintf (stream, "%s %s %s %d %1.4f %d %g %g %g %g pv\n", Y->row_names[j], chrom->name, snps[i]->rs, snps[i]->position, f, bayes->X->size1, a, r2, pval, tbf);
						fflush (stream);
					}
				}
			}
						
		} // snp
		
		if (chrom->logger)
		{
			gdl_progress_bar (chrom->logger, 25, j, (to-from));
		}
				
	} // feature
	
	// clean workspace
	GDL_FREE (popf);
	GDL_FREE (bf);
	gdl_bayreg_workspace_free (bayes);
	
	return GDL_SUCCESS;
}

void
gdl_igdb_reg_permut (gdl_bayreg_workspace * bayes, gdl_igdb_chromosome * chrom, const gdl_rng * rng, const size_t within)
{
	size_t i, j, k, l, p = 0;
	double t;
	
	if (within)
	{
		for (k = i = 0; i < chrom->npop; i++)
		{
			for(j = chrom->pop_sizes[i]-1; j > 0; j--)
			{
				l = gdl_rng_uniform_int (rng, j+1);
				t = MATRIX_GET(bayes->Y, k+l, p);
				MATRIX_SET(bayes->Y, k+l, p, MATRIX_GET(bayes->Y, k+j, p));
				MATRIX_SET(bayes->Y, k+j, p, t);
			}
			k += chrom->pop_sizes[i];
		}
	}
	else
	{
		for(j = chrom->nindiv-1; j > 0; j--)
		{
			l = (j + 1) * gdl_rng_uniform (rng);
			t = MATRIX_GET(bayes->Y, l, p);
			MATRIX_SET(bayes->Y, l, p, MATRIX_GET(bayes->Y, j, p));
			MATRIX_SET(bayes->Y, j, p, t);
		}	
	}

}

/**.
 * Regress the phenotypes Y[1..Y->size1]
 * individuals SNPs[from..to] and output
 * only SNPs fullfilling the criteria  
 */
int
gdl_igdb_chromosome_reg_permut (gdl_igdb_chromosome * chrom,
	                             gdl_bayreg_model * model,
	                             const gdl_boolean qnorm,
	                             size_t * rm_individual[],
	                             const gdl_feature_table * feature_table,
	                             const gdl_labeled_matrix * Y,
	                             const int y_from,
	                             const int y_to,
	                             const gdl_rng * rng,
	                             const size_t permut,
	                             gdl_igdb_reg_best_buffer * best,
	                             const double bf_min,
	                             const double pv_max,
	                             FILE * stream)
{
	const size_t G    = chrom->ftr_data->nfeature;
	const size_t P    = chrom->ploidy;
	const size_t N    = chrom->nindiv;
	const size_t POP  = chrom->npop;
	const size_t NSNP = chrom->snp_data->chrom->size;
	size_t i, j, k, l, msize, from, to, npop, nbf, npv;
	double * bf, * popf, tbf, f, pval, a, r2, chi, max_bf, min_pv;
	gdl_snp ** snps;
	gdl_bayreg_workspace * bayes;
	
	snps = chrom->snp_data->chrom->snps;
	
	msize = gdl_bayreg_model_bf_storage_size (model);
	
	bayes = gdl_bayreg_workspace_alloc (N, POP);
	
	popf = GDL_CALLOC (double, POP);
	
	bf = GDL_CALLOC (double, msize);
	
	from = (y_from == -1) ? 0 : y_from;
	
	to   = (y_to == -1) ? Y->X->size1 : y_to;
	
	if (chrom->logger)
	{
		fprintf (chrom->logger, ">%s |nfeature=%d|nsnp=%d|npop=%d|nindiv=%d|\n", chrom->name, to-from, NSNP, POP, N);
		fflush(chrom->logger);
	}
	
	for(j = from; j <  to; j++)
	{
		if (feature_table && !gdl_feature_table_is_inside (feature_table, chrom->name, Y->row_names[j]))
		{
			continue;
		}
		
		gdl_igdb_reg_feature_init (bayes, chrom, Y->X, j, qnorm);
		
		for(k = 0; k < permut; k++)
		{
			gdl_igdb_reg_permut (bayes, chrom, rng, qnorm);
			
			max_bf = 0;
			min_pv = 1.0;
			nbf = npv = 0;
				
			for (i = 0; i < NSNP; i++)
			{
				npop = gdl_igdb_reg_snp_init (bayes, chrom->snp_data, N, i, popf, &f, rm_individual);
					
				if (npop)
				{
					gdl_bayreg_model_single_bf (model, bayes, bf);
					gdl_igdb_reg_perform_snp_add (bayes, &a, &r2, &pval);
					tbf = gdl_bayreg_model_average_bf (model, bf, 0);
					if (best)
					{
						gdl_igdb_reg_best_buffer_test_bf (best, j, tbf, 'p');
						gdl_igdb_reg_best_buffer_test_pv (best, j, pval, 'p');
					}
					if (stream)
					{
						if (tbf > bf_min)
						{
							fprintf (stream, "bf %g\n", tbf);
							fflush (stream);
						}
						if (pval < pv_max)
						{
							fprintf (stream, "pv %g\n", pval);
							fflush (stream);
						}
					}
				} // snp
							
			} // permut
			
		} // feature
		
		if (chrom->logger)
		{
			gdl_progress_bar (chrom->logger, 25, j, (to-from));
		}
				
	} // feature
	
	// clean workspace
	GDL_FREE (popf);
	GDL_FREE (bf);
	gdl_bayreg_workspace_free (bayes);
	
	return GDL_SUCCESS;
} 
 
