/*
 *  igdb/looper.c
 *
 *  $Author: baptiste $, $Date: 2008-05-13 15:33:47 $, $Version$
 *
 *  Libgdl : a C library for statistical genetics
 *
 *  Copyright (C) 2009 Jean-Baptiste Veyrieras, University of Chicago, USA.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
 */

#include <gdl/gdl_common.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_string.h>
#include <gdl/gdl_math.h>
#include <gdl/gdl_snp_genome.h>
#include <gdl/gdl_expr_genome.h>
#include <gdl/gdl_expr_feature.h>
#include <gdl/gdl_expr_feature_cis.h>
#include <gdl/gdl_igdb_chromosome.h>
#include <gdl/gdl_igdb_genome.h>
#include <gdl/gdl_igdb_cis_annot.h>
#include <gdl/gdl_igdb_looper.h>

gdl_igdb_cis_looper *
gdl_igdb_cis_looper_alloc (const gdl_igdb_genome * db, const gdl_igdb_cis_annot * annot, const gdl_feature_table * gene_set, const gdl_feature_table * gene_rm, const gdl_snp_table * snp_rm, const long window, FILE * logger)
{
	gdl_igdb_cis_looper * l;

	l = GDL_CALLOC (gdl_igdb_cis_looper, 1);

	l->db  = db;
	l->gene_set = gene_set;
	l->gene_rm = gene_rm;
	l->snp_rm   = snp_rm;
	l->annot    = annot;
	l->window   = window;
	l->logger   = logger;

	return l;
}

void
gdl_igdb_cis_looper_free (gdl_igdb_cis_looper * l)
{
	if (l)
	{
		GDL_FREE (l);
	}
}

int
gdl_igdb_cis_looper_init_chromosome_loop (gdl_igdb_cis_looper * hm, const size_t c)
{
	if (hm->gene_set && !gdl_feature_table_has_chromosome (hm->gene_set, hm->db->chroms[c]))
	{
		//(hm->_chrom_idx)++;
		return 0;
	}

	hm->_chrom  = gdl_igdb_genome_get (hm->db, c);

	if (hm->_chrom->snp_data)
	{
		hm->_snps   = hm->_chrom->snp_data->chrom->snps;

		gdl_igdb_chromosome_flag_snp_on_ftr_ends (hm->_chrom);

		gdl_igdb_chromosome_flag_snp_on_ftr_window_end (hm->_chrom, hm->window);
	}

	if (hm->logger)
	{
		fprintf (hm->logger, ">%s |nfeature=%d|\n", hm->_chrom->name, (hm->_chrom->ftr_data) ? hm->_chrom->ftr_data->nfeature : 0);
		fflush (hm->logger);
	}

	if (hm->annot) // Load annotation tracks
	{
		gdl_igdb_cis_annot_load (hm->annot, hm->_chrom);
	}

	return (hm->_chrom->ftr_data) ? hm->_chrom->ftr_data->nfeature : 0;
}

int
gdl_igdb_cis_looper_end_chromosome_loop (gdl_igdb_cis_looper * hm, const size_t c)
{
	if (hm->annot)
	{
		gdl_igdb_cis_annot_unload (hm->annot);
	}
	gdl_igdb_chromosome_free (hm->_chrom);
	hm->_chrom  = 0;
}

int
gdl_igdb_cis_looper_init_feature_loop (gdl_igdb_cis_looper * hm, const size_t g)
{
	hm->_feature      = hm->_chrom->ftr_data->features[g];
	hm->_feature_idx  = g;

	if (hm->gene_set && !gdl_feature_table_is_inside (hm->gene_set, hm->_chrom->name, hm->_feature->name))
	{
		return 0;
	}

	if (hm->gene_rm && gdl_feature_table_is_inside (hm->gene_rm, hm->_chrom->name, hm->_feature->name))
	{
		return 0;
	}

	if (hm->annot) // focus on gene cis-region
	{
		size_t ne, ns;
		const long * fs = (hm->_feature->T->get_starts)(hm->_chrom->exp_data, hm->_feature->idx, &ns);
		const long * fe = (hm->_feature->T->get_ends)(hm->_chrom->exp_data, hm->_feature->idx, &ne);
		gdl_igdb_cis_annot_focus (hm->annot, fs[0] - hm->window, fe[ne-1] + hm->window, hm->_thread_idx);
	}

	if (hm->_chrom->ftr_cis)
	{
		hm->_cis_feature = hm->_chrom->ftr_cis->features[g];
	}

	return 1;
}

void
gdl_igdb_cis_looper_end_feature_loop (gdl_igdb_cis_looper * hm, const size_t g)
{
	if (hm->logger)
	{
		gdl_progress_bar (hm->logger, (hm->_chrom->ftr_data->nfeature > 25) ? 25 : hm->_chrom->ftr_data->nfeature - 1, g, hm->_chrom->ftr_data->nfeature);
		fflush (hm->logger);
	}
	if (hm->annot) // focus on gene cis-region
	{
		gdl_igdb_cis_annot_unfocus (hm->annot, hm->_thread_idx);
	}
	hm->_feature     = 0;
	hm->_cis_feature = 0;
}

size_t
gdl_igdb_cis_looper_snp_window_start (const gdl_igdb_cis_looper * hm)
{
	if (hm->_cis_feature)
	{
		return GDL_MAX (hm->_cis_feature->snp_from, hm->_chrom->ftr_snp_up[hm->_feature_idx]) - 1;
	}
	else
	{
		return hm->_chrom->ftr_snp_up[hm->_feature_idx];
	}
}

size_t
gdl_igdb_cis_looper_snp_window_end (const gdl_igdb_cis_looper * hm)
{
	if (hm->_cis_feature)
	{
		return GDL_MIN(hm->_chrom->snp_data->chrom->size-1, GDL_MIN (hm->_cis_feature->snp_to, hm->_chrom->ftr_snp_down[hm->_feature_idx]));
	}
	else
	{
		return GDL_MIN(hm->_chrom->snp_data->chrom->size-1, hm->_chrom->ftr_snp_down[hm->_feature_idx]);
	}
}

gdl_boolean
gdl_igdb_cis_looper_snp_next (gdl_igdb_cis_looper * hm, const size_t l)
{
	gdl_snp * snp = hm->_snps[l];

	if (snp->ignore=='y' && !hm->unignore)
	{
		return gdl_false;
	}

	if (hm->snp_rm && gdl_snp_table_is_inside (hm->snp_rm, hm->_chrom->name, hm->_feature->name, snp->rs))
	{
		return gdl_false;
	}

	size_t ne, ns;
	const long * fs = (hm->_feature->T->get_starts)(hm->_chrom->exp_data, hm->_feature->idx, &ns);
	const long * fe = (hm->_feature->T->get_ends)(hm->_chrom->exp_data, hm->_feature->idx, &ne);
	long tss_pos    = fs[0] - snp->position;
	if (tss_pos > hm->window)
	{
		return gdl_false;
	}
	long tes_pos = snp->position-fe[ne-1];
	if (tes_pos > hm->window)
	{
		return gdl_false;
	}
	if (hm->_cis_feature)
	{
		hm->_cis_snp = hm->_cis_feature->snps[l-hm->_cis_feature->snp_from];
		if (hm->_cis_snp==0)
		{
			return gdl_false;
		}
	}
	hm->_snp  = snp;
	return gdl_true;
}

//=============================================================================================
//
// Multi-thread Interface
//
//=============================================================================================

gdl_igdb_cis_looperth *
gdl_igdb_cis_looperth_alloc (const size_t nthread, const gdl_igdb_genome * db, gdl_igdb_cis_annot * annot, const gdl_feature_table * gene_set, const gdl_feature_table * gene_rm, const gdl_snp_table * snp_rm, const long window)
{
	size_t i;
	gdl_igdb_cis_looperth * l;

	l  = GDL_CALLOC (gdl_igdb_cis_looperth, 1);
	l->nthread  = nthread;
	l->looperth = GDL_CALLOC (gdl_igdb_cis_looper *, nthread);
	l->annot    = annot;
	gdl_igdb_cis_annot_mth (l->annot, nthread);
	for(i = 0; i < nthread; i++)
	{
		l->looperth[i] = gdl_igdb_cis_looper_alloc (db, annot, gene_set, gene_rm, snp_rm, window, 0);
		l->looperth[i]->_thread_idx = i;
	}

	return l;
}

void
gdl_igdb_cis_looperth_free (gdl_igdb_cis_looperth * l)
{
	if (l)
	{
		size_t i;
		for(i = 0; i <  l->nthread; i++)
		{
			gdl_igdb_cis_looper_free (l->looperth[i]);
		}
		GDL_FREE (l->looperth);
		GDL_FREE (l);
	}
}

int
gdl_igdb_cis_looper_init_chromosome_loopth (gdl_igdb_cis_looperth * hm, const size_t t, const size_t c)
{
	return gdl_igdb_cis_looper_init_chromosome_loop(hm->looperth[t], c);
}

int
gdl_igdb_cis_looper_end_chromosome_loopth (gdl_igdb_cis_looperth * hm, const size_t t, const size_t c)
{
	return gdl_igdb_cis_looper_end_chromosome_loop(hm->looperth[t], c);
}

int
gdl_igdb_cis_looper_init_feature_loopth (gdl_igdb_cis_looperth * hm, const size_t t, const size_t g)
{
	return gdl_igdb_cis_looper_init_feature_loop(hm->looperth[t], g);
}

void
gdl_igdb_cis_looper_end_feature_loopth (gdl_igdb_cis_looperth * hm, const size_t t, const size_t g)
{
	return gdl_igdb_cis_looper_end_feature_loop(hm->looperth[t], g);
}

size_t
gdl_igdb_cis_looper_snp_window_startth (const gdl_igdb_cis_looperth * hm, const size_t t)
{
	return gdl_igdb_cis_looper_snp_window_start(hm->looperth[t]);
}

size_t
gdl_igdb_cis_looper_snp_window_endth (const gdl_igdb_cis_looperth * hm, const size_t t)
{
	return gdl_igdb_cis_looper_snp_window_end(hm->looperth[t]);
}

gdl_boolean
gdl_igdb_cis_looper_snp_nextth (gdl_igdb_cis_looperth * hm, const size_t t, const size_t l)
{
	return gdl_igdb_cis_looper_snp_next(hm->looperth[t], l);
}
