/* matrix/labeled.c
 * 
 * Copyright (C) 2009, Jean-Baptiste Veyrieras
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <gdl/gdl_common.h>
#include <gdl/gdl_types.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_string.h>
#include <gdl/gdl_matrix.h>
#include <gdl/gdl_labeled_matrix.h>

gdl_labeled_matrix * 
gdl_labeled_matrix_alloc (const size_t N, const size_t M)
{
	gdl_labeled_matrix * m;
	
	m = GDL_CALLOC (gdl_labeled_matrix, 1);
	
	m->X = gdl_matrix_calloc (N, M);
	m->row_names = GDL_CALLOC (gdl_string *, N);
	m->col_names = GDL_CALLOC (gdl_string *, M);
	
	return m;
}

void
gdl_labeled_matrix_free (gdl_labeled_matrix * m)
{
	if (m)
	{
		size_t i;
		for(i = 0; i < m->X->size1; i++)
			gdl_string_free (m->row_names[i]);
		for(i = 0; i < m->X->size2; i++)
			gdl_string_free (m->col_names[i]);
		GDL_FREE (m->row_names);
		GDL_FREE (m->col_names);
		gdl_matrix_free (m->X);
		GDL_FREE (m);	
	}
}

int
gdl_labeled_matrix_fwrite (FILE * stream, const gdl_labeled_matrix * m)
{
	if (stream && m)
	{
		size_t i;
		int status;
		
		status = gdl_matrix_fwrite (stream, m->X);
		GDL_FWRITE_STATUS (status, GDL_SUCCESS);
		for(i = 0; i < m->X->size1; i++)
		{
			status = gdl_string_fwrite (stream, m->row_names[i]);
			GDL_FWRITE_STATUS (status, GDL_SUCCESS);
		}
		for(i = 0; i < m->X->size2; i++)
		{
			status = gdl_string_fwrite (stream, m->col_names[i]);
			GDL_FWRITE_STATUS (status, GDL_SUCCESS);
		}
				
		return GDL_SUCCESS;
	}
	return GDL_EINVAL;
}

gdl_labeled_matrix * 
gdl_labeled_matrix_fread (FILE * stream)
{
	if (stream)
	{
		size_t i;
		int status;
		gdl_labeled_matrix * m;
		
		m = GDL_CALLOC (gdl_labeled_matrix, 1);
		
		m->X = gdl_matrix_fread (stream);
		GDL_FREAD_STATUS (m->X!=0, 1);
		
		m->row_names = GDL_CALLOC (gdl_string *, m->X->size1);
		m->col_names = GDL_CALLOC (gdl_string *, m->X->size2);
		 
		for(i = 0; i < m->X->size1; i++)
		{
			m->row_names[i] = gdl_string_fread (stream);
			GDL_FWRITE_STATUS (m->row_names[i]!=0, 1);
		}
		for(i = 0; i < m->X->size2; i++)
		{
			m->col_names[i] = gdl_string_fread (stream);
			GDL_FWRITE_STATUS (m->col_names[i]!=0, 1);
		}
				
		return m;
	}
	return 0;
}

int
gdl_labeled_matrix_fprintf (FILE * stream, const gdl_labeled_matrix * m)
{
	size_t i, j;
	
	fprintf (stream, "%s", m->col_names[0]);
	for(i = 1; i < m->X->size2; i++)
	{
		fprintf (stream, " %s", m->col_names[i]);	
	}
	fprintf (stream, "\n");
	fflush (stream);
	for(i = 0; i < m->X->size1; i++)
	{
		fprintf (stream, "%s", m->row_names[i]);
		for(j = 0; j < m->X->size2; j++)
		{
			fprintf (stream, " %g", gdl_matrix_get (m->X, i, j));
		}
		fprintf (stream, "\n");
		fflush (stream);
	}
}
