/*
 *  igdb/genome.c
 *
 *  $Author: baptiste $, $Date: 2008-05-13 15:33:47 $, $Version$
 *
 *  Libgdl : a C library for statistical genetics
 *
 *  Copyright (C) 2009 Jean-Baptiste Veyrieras, University of Chicago, USA.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
 */

#include <stddef.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_io.h>
#include <gdl/gdl_string.h>
#include <gdl/gdl_hash.h>
#include <gdl/gdl_table.h>
#include <gdl/gdl_labeled_matrix.h>
#include <gdl/gdl_snp_genome.h>
#include <gdl/gdl_expr_genome.h>
#include <gdl/gdl_expr_feature.h>
#include <gdl/gdl_expr_feature_cis.h>
#include <gdl/gdl_igdb_genome.h>

// for clarity
typedef struct
{
	size_t npop;
	gdl_string ** pops;
	size_t nchrom;
	gdl_string ** chroms;
	size_t ploidy;
	size_t snptype;
	gdl_string * snpdir;
	gdl_string * snpDB;
	unsigned char snpna;
	gdl_string * snpgmapdir;
	gdl_string * expdir;
	gdl_string * expDB;
	gdl_string * prbanodir;
	gdl_string * txtabdir;
	gdl_string * expna;
	gdl_string * sampinf;
	gdl_string * sampna;
	gdl_string * output;
	long prbwindow;
} gdl_igdb_genome_config;

static int
_gdl_igdb_genome_read (const gdl_string * filename,
		               gdl_igdb_genome * g,
		               FILE * logger)
{
	size_t i,j,n;
	gdl_string * key, * value, * line = 0;
	gdl_igdb_genome_config *Cptr, C;
	FILE * stream;
	DIR * dir;

	Cptr = GDL_CALLOC (gdl_igdb_genome_config, 1);
	C = *Cptr;

	C.snpna  = '?';
	C.expna  = gdl_string_sprintf("NA");
	C.sampna = gdl_string_sprintf("NA");

	stream = gdl_fileopen (filename, "r");
	while (gdl_getline (&line, &n, stream)!=-1)
	{
		if (line[0]!='#')
		{
			i=j=0;
			key   = gdl_string_next_token (line, n, &i, &j);
			value = gdl_string_next_token (line, n, &i, &j);
			if (!strcmp(key, "Pop")) // the name of the populations
			{
				C.pops = gdl_string_split (value, ",", &(C.npop));
				gdl_string_free (value);
			}
			else if (!strcmp(key, "Ploidy")) // 1 or 2
			{
				C.ploidy = (size_t)atol(value);
				if (C.ploidy > 2 || C.ploidy == 0)
				{
					GDL_ERROR_VAL ("Ploidy must be 1 or 2", GDL_EINVAL, GDL_EINVAL);
				}
				gdl_string_free (value);
			}
			if (!strcmp(key, "SNPDir")) // the location of the population directories
			{
				C.snpdir = value;
				DIR * testDir = opendir (C.snpdir);
				if (!testDir)
				{
					GDL_ERROR_VAL (gdl_string_sprintf("SNP Directory (SNPDir=%s) is not accessible", value), GDL_EINVAL, GDL_EINVAL);
				}
			}
			else if (!strcmp(key, "SNPDB")) // Existing SNP database
			{
				C.snpDB = value;
				if (!gdl_isfile (value))
				{
					GDL_ERROR_VAL (gdl_string_sprintf("SNP database (SNPDB=%s) is not accessible", value), GDL_EINVAL, GDL_EINVAL);
				}
			}
			else if (!strcmp(key, "SNPMissing")) // missing token for SNP
			{
				C.snpna = (unsigned char) value[0];
				if (strlen(value) > 1)
				{
					GDL_ERROR_VAL ("Missing code for SNP must be a single character, not a string", GDL_EINVAL, GDL_EINVAL);
				}
				gdl_string_free (value);
			}
			else if (!strcmp(key, "SNPGMapDir")) // SNP genetic map
			{
				C.snpgmapdir = value;
				DIR * testDir = opendir (C.snpgmapdir);
				if (!testDir)
				{
					GDL_ERROR_VAL (gdl_string_sprintf("Genetic Map Directory (SNPGMapDir=%s) is not accessible", value), GDL_EINVAL, GDL_EINVAL);
				}
			}
			else if (!strcmp(key, "SNPDataType")) // SNP haplotype / genotype / imputed
			{
				if (!strncmp(value, "1", 1))
				{
					C.snptype = 0;
				}
				else if (!strncmp(value, "2", 1))
				{
					C.snptype = 1;
				}
				else if (!strncmp(value, "3", 1))
				{
					C.snptype = 2;
				}
				else if (!strcmp(value, "haplotype"))
				{
					C.snptype = 0;
				}
				else if (!strcmp(value, "genotype"))
				{
					C.snptype = 1;
				}
				else if (!strcmp(value, "imputed"))
				{
					C.snptype = 2;
				}
				else
				{
					GDL_ERROR_VAL (gdl_string_sprintf("Unknown SNP Data Type %s", value), GDL_EINVAL, GDL_EINVAL);
				}
			}
			else if (!strcmp(key, "ExpDir")) // expression data file for each population
			{
				C.expdir = value;
				DIR * testDir = opendir (value);
				if (!testDir)
				{
					GDL_ERROR_VAL (gdl_string_sprintf("Expression Data Directory (ExpDir=%s) is not accessible", value), GDL_EINVAL, GDL_EINVAL);
				}
			}
			else if (!strcmp(key, "ExpDB")) // Existing expression database
			{
				C.expDB = value;
				if (!gdl_isfile (value))
				{
					GDL_ERROR_VAL (gdl_string_sprintf("Expression database (ExpDB=%s) is not accessible", value), GDL_EINVAL, GDL_EINVAL);
				}
			}
			else if (!strcmp(key, "PrbAnnotDir")) // Probe annotation file
			{
				C.prbanodir = value;
				DIR * testDir = opendir (value);
				if (!testDir)
				{
					GDL_ERROR_VAL (gdl_string_sprintf("Probe Annotation Directory (PrbAnnotDir=%s) is not accessible", value), GDL_EINVAL, GDL_EINVAL);
				}
			}
			else if (!strcmp(key, "PrbTSSWindow")) // Probe annotation file
			{
				C.prbwindow = atol(value);
				if (C.prbwindow < 0)
				{
					GDL_ERROR_VAL ("Invalid Probe TSS Window: Must be a positive integer", GDL_EINVAL, GDL_EINVAL);
				}
				gdl_string_free (value);
			}
			else if (!strcmp(key, "TxTableDir")) // The Transcript Table
			{
				C.txtabdir = value;
				DIR * testDir = opendir (value);
				if (!testDir)
				{
					GDL_ERROR_VAL (gdl_string_sprintf("Transcript Annotation Directory (TxTableDir=%s) is not accessible", value), GDL_EINVAL, GDL_EINVAL);
				}
			}
			else if (!strcmp(key, "ExpMissing")) // missing token for expression values
			{
				C.expna = value;
			}
			else if (!strcmp(key, "SampleInfo")) // sample information file
			{
				C.sampinf = value;
				FILE * testFile = gdl_fileopen (value, "r");
				if (!testFile)
				{
					GDL_ERROR_VAL (gdl_string_sprintf("Sample Information File (SampleInfo=%s) is not accessible", value), GDL_EINVAL, GDL_EINVAL);
				}
				else
				{
					gdl_fileclose (value, testFile);
				}
			}
			else if (!strcmp(key, "SampleInfoMissing")) // sample information missing string
			{
				C.sampna = value;
			}
			else if (!strcmp(key, "OutputStem")) // sample information file
			{
				C.output = value;
			}
			else if (!strcmp(key, "Chromosome")) // the chromosome to look at
			{
				C.chroms = gdl_string_split (value, ",", &(C.nchrom));
				gdl_string_free (value);
			}
			gdl_string_free (key);
		}
		gdl_string_free (line);
		line=0;
	}

	gdl_fileclose (filename, stream);

	if (C.nchrom == 0)
	{
		GDL_ERROR_VAL ("You must provide the name of the chromosomes (e.g. Chromosome chr1,chr2,...)", GDL_EINVAL, GDL_EINVAL);
	}

	if (C.npop == 0)
	{
		// Assume a single population
		if (logger)
		{
			fprintf (logger, "[ WARNING ] No population folder names were provided: ASSUME A SINGLE POPULATION\n");
			fflush (logger);
		}
		C.npop    = 1;
		C.pops    = GDL_MALLOC (gdl_string *, 1);
		C.pops[0] = gdl_string_sprintf("");
	}

	if ((dir=opendir(C.output))==0)
	{
		if (logger)
		{
			fprintf (logger, "\n--\nCreate the output directory %s\n--\n", C.output);
			fflush (logger);
		}
		if (mkdir (C.output, S_IRWXU)!=0)
		{
			GDL_ERROR_VAL (gdl_string_sprintf("Unable to create output directory %s", C.output), GDL_EINVAL, GDL_EINVAL);
		}
	}

	size_t nsample = 0;

	if (C.snpDB)
	{
		if (logger)
		{
			fprintf (logger, "\n--\nLoad the existing SNP genotype database\n--\n");
			fflush (logger);
		}
		stream = gdl_fileopen (C.snpDB, "r");
		g->s_genome = gdl_snp_genome_fread (stream);
		gdl_fileclose(C.snpDB, stream);
		if (!g->s_genome)
		{
			GDL_ERROR_VAL ("Unable to load existing SNP database", GDL_EINVAL, 0);
		}
		if (g->s_genome->nchrom != C.nchrom)
		{
			GDL_ERROR_VAL ("Mismatched number of chromosomes for the existing SNP database", GDL_EINVAL, 0);
		}
	}
	else if (C.snpdir)
	{
		if (logger)
		{
			fprintf (logger, "\n--\nLoad the SNP database\n--\n");
			fflush (logger);
		}
		gdl_snp_data_format * F = gdl_snp_data_format_alloc (C.ploidy);

		F->na_char = C.snpna;
		F->haplo   = (C.snptype==0) ? 1 : 0;
		F->imputed = (C.snptype==2) ? 1 : 0;

		g->s_genome = gdl_snp_genome_alloc2 (C.snpdir, C.pops, C.npop, C.chroms, C.nchrom, F, C.output);

		nsample = g->s_genome->_nsample;

		if (C.snpgmapdir)
		{
			gdl_snp_genome_gmap (g->s_genome, C.snpgmapdir);
		}

		if (logger)
		{
			fprintf (logger, "\n--\nSNP database [ OK ] \n--\n");
			fflush (logger);
		}

	}

	if (C.expDB)
	{
		if (logger)
		{
			fprintf (logger, "\n--\nLoad the existing expression database\n--\n");
			fflush (logger);
		}
		stream = gdl_fileopen (C.expDB, "r");
		g->e_genome = gdl_expr_genome_fread (stream);
		gdl_fileclose(C.expDB, stream);
		if (!g->e_genome)
		{
			GDL_ERROR_VAL ("Unable to load existing expression database", GDL_EINVAL, 0);
		}
		if (g->e_genome->nchrom != C.nchrom)
		{
			GDL_ERROR_VAL ("Mismatched number of chromosomes for the existing expression database", GDL_EINVAL, 0);
		}
		// SHOULD CHECK HERR IF NUMBER OF POP, etc.. is consistent
	}
	else if (C.expdir)
	{
		if (logger)
		{
			fprintf (logger, "\n--\nLoad the Expression database\n--\n");
			fflush (logger);
		}

		if (C.prbanodir == 0)
		{
			if (C.txtabdir == 0)
			{
				GDL_ERROR_VAL ("When no probe annotation is provided, you must provide a transcript table", GDL_EINVAL, GDL_EINVAL);
			}

			g->e_genome = gdl_expr_genome_alloc3 (C.expdir, C.pops, C.npop, C.chroms, C.nchrom, C.txtabdir, C.output, C.expna, logger, stderr);

			if (nsample && g->e_genome->_nsample != nsample)
			{
				GDL_WARNING (gdl_string_sprintf("SNP data contains %d samples and expression data %d samples", nsample, g->e_genome->_nsample), GDL_EINVAL);
			}
			else
			{
				nsample = g->e_genome->_nsample;
			}

		}
		else if (C.prbwindow > 0)
		{
			g->e_genome = gdl_expr_genome_alloc4 (C.expdir, C.pops, C.npop, C.chroms, C.nchrom, C.prbanodir, C.txtabdir, C.output, C.expna, C.prbwindow, logger, stderr);
			if (nsample && g->e_genome->_nsample != nsample)
			{
				GDL_WARNING (gdl_string_sprintf("SNP data contains %d samples and expression data %d samples", nsample, g->e_genome->_nsample), GDL_EINVAL);
			}
			else
			{
				nsample = g->e_genome->_nsample;
			}
		}
		else
		{
			g->e_genome = gdl_expr_genome_alloc2 (C.expdir, C.pops, C.npop, C.chroms, C.nchrom, C.prbanodir, C.txtabdir, C.output, C.expna, logger, stderr);
			if (nsample && g->e_genome->_nsample != nsample)
			{
				GDL_WARNING (gdl_string_sprintf("SNP data contains %d samples and expression data %d samples", nsample, g->e_genome->_nsample), GDL_EINVAL);
			}
			else
			{
				nsample = g->e_genome->_nsample;
			}
		}

		if (logger)
		{
			fprintf (logger, "\n--\nExpression database [ OK ] \n--\n");
			fflush (logger);
		}
	}

	if (C.sampinf)
	{
		if (logger)
		{
			fprintf (logger, "\n--\nLoad the sample information table\n--\n");
			fflush (logger);
		}

		stream = gdl_fileopen (C.sampinf, "r");
		g->sample_info = gdl_table_fscanf (stream, C.sampna);
		gdl_fileclose (C.sampinf, stream);

		// check if as many samples as in the snp x expression database
		if (g->sample_info->nrow != nsample)
		{
			GDL_WARNING (gdl_string_sprintf ("SampleInfo file has %d samples when %d are expected\n", g->sample_info->nrow, nsample), GDL_EINVAL);
		}

		if (logger)
		{
			fprintf (logger, "\n--\nSample information table [ OK ] \n--\n");
			fflush (logger);
		}
	}

	g->nchrom = C.nchrom;
	g->chroms = C.chroms;
	g->dbdir  = C.output;
	g->npop   = C.npop;
	g->pop_names = C.pops;

	// Clean the rest of C

	return GDL_SUCCESS;
}

gdl_igdb_genome *
gdl_igdb_genome_alloc (const gdl_string * config_file)
{
	gdl_igdb_genome * g;

	g = GDL_CALLOC (gdl_igdb_genome, 1);

	if (_gdl_igdb_genome_read (config_file, g, stderr) != GDL_SUCCESS)
	{
		GDL_ERROR_VAL ("Unable to create the database", GDL_FAILURE, 0);
	}

	return g;
}

void
gdl_igdb_genome_free (gdl_igdb_genome * v)
{
	if (v)
	{
		size_t i;
		for (i = 0; i < v->nchrom; i++)
		{
			gdl_string_free (v->chroms[i]);
		}
		GDL_FREE (v->chroms);
		for (i = 0; i < v->npop; i++)
		{
			gdl_string_free (v->pop_names[i]);
		}
		GDL_FREE (v->pop_names);
		gdl_string_free (v->dbdir);
		gdl_table_free (v->sample_info);
		GDL_FREE (v);
	}
}

gdl_igdb_chromosome *
gdl_igdb_genome_get (const gdl_igdb_genome * g, size_t i)
{
	size_t j;
	gdl_igdb_chromosome * c = gdl_igdb_chromosome_alloc (g->chroms[i], i);

	if (g->s_genome)
	{
		c->snp_data = gdl_snp_genome_get (g->s_genome, i);
		c->ploidy    = c->snp_data->pops[0]->P;
	}
	if (g->e_genome)
	{
		c->exp_data  = gdl_expr_genome_get (g->e_genome, i);
		c->npop      = c->exp_data->npop;
		c->pop_sizes = c->exp_data->pop_sizes;
		c->nindiv    = 0;
		for(j = 0; j < c->npop; j++)
		{
			c->nindiv += c->pop_sizes[j];
		}
	}
	if (c->exp_data != 0 && g->f_genome != 0)
	{
		c->ftr_data = gdl_expr_feature_genome_get (g->f_genome, i, c->exp_data);
		c->nfeature = c->ftr_data->nfeature;
	}
	if (c->exp_data != 0 && g->r_genome != 0)
	{
		c->ftr_cis  = gdl_expr_feature_cis_genome_get (g->r_genome, i);
	}

	return c;
}

gdl_igdb_chromosome *
gdl_igdb_genome_lookup (const gdl_igdb_genome * g, const gdl_string * name)
{
	size_t i;

	for(i = 0; i < g->nchrom; i++)
	{
		if (!strcmp(g->chroms[i], name))
		{
			return gdl_igdb_genome_get (g, i);
		}
	}

	return NULL;
}

int
gdl_igdb_genome_set (const gdl_igdb_genome * g, size_t i, gdl_igdb_chromosome * c)
{
	if (c->snp_data)
	{
		gdl_snp_genome_set (g->s_genome, i, c->snp_data);
	}
	if (c->exp_data)
	{
		gdl_expr_genome_set (g->e_genome, i, c->exp_data);
	}

	return GDL_SUCCESS;
}

gdl_igdb_genome *
gdl_igdb_genome_fread (FILE * stream)
{
	if (stream)
	{
		int i, status;
		unsigned char has;
		gdl_igdb_genome * v;

		v = GDL_CALLOC (gdl_igdb_genome, 1);

		v->dbdir = gdl_string_fread (stream);
		GDL_FREAD_STATUS (v->dbdir!=0, 1);

		status = fread (&v->nchrom, sizeof(size_t), 1, stream);
		GDL_FREAD_STATUS (status, 1);

		v->chroms = GDL_MALLOC (gdl_string *, v->nchrom);
		for (i = 0; i < v->nchrom; i++)
		{
			v->chroms[i] = gdl_string_fread (stream);
			GDL_FREAD_STATUS (v->chroms[i]!=0, 1);
		}

		status = fread (&has, sizeof(unsigned char), 1, stream);
		GDL_FREAD_STATUS (status, 1);

		if (has == 'y')
		{
			v->s_genome = gdl_snp_genome_fread (stream);
			GDL_FREAD_STATUS (v->s_genome != 0, 1);
		}

		status = fread (&has, sizeof(unsigned char), 1, stream);
		GDL_FREAD_STATUS (status, 1);

		if (has == 'y')
		{
			v->e_genome = gdl_expr_genome_fread (stream);
			GDL_FREAD_STATUS (v->e_genome != 0, 1);
		}

		status = fread (&has, sizeof(unsigned char), 1, stream);
		GDL_FREAD_STATUS (status, 1);

		if (has == 'y')
		{
			v->sample_info = gdl_table_fread (stream);
			GDL_FREAD_STATUS (v->sample_info != 0, 1);
		}

		if (!gdl_eoff (stream))
		{
			status = fread (&(v->npop), sizeof(size_t), 1, stream);
			GDL_FREAD_STATUS (status, 1);
			v->pop_names = GDL_MALLOC (gdl_string *, v->npop);
			for(i = 0; i < v->npop; i++)
			{
				v->pop_names[i] = gdl_string_fread (stream);
				GDL_FREAD_STATUS (v->pop_names[i]!=0, 1);
			}
		}

		return v;
	}
	return 0;
}

int
gdl_igdb_genome_fwrite (FILE * stream, const gdl_igdb_genome * v)
{
	if (stream && v)
	{
		int status;
		size_t i;
		unsigned char has;

		status = gdl_string_fwrite (stream, v->dbdir);
		GDL_FWRITE_STATUS (status, GDL_SUCCESS);

		status = fwrite (&v->nchrom, sizeof(size_t), 1, stream);
		GDL_FWRITE_STATUS (status, 1);

		for (i = 0; i < v->nchrom; i++)
		{
			status = gdl_string_fwrite (stream, v->chroms[i]);
			GDL_FWRITE_STATUS (status, GDL_SUCCESS);
		}

		has = (v->s_genome) ? 'y' : 'n';
		status = fwrite (&has, sizeof(unsigned char), 1, stream);
		GDL_FWRITE_STATUS (status, 1);

		if (v->s_genome)
		{
			status = gdl_snp_genome_fwrite (stream, v->s_genome);
			GDL_FWRITE_STATUS (status, GDL_SUCCESS);
		}

		has = (v->e_genome) ? 'y' : 'n';
		status = fwrite (&has, sizeof(unsigned char), 1, stream);
		GDL_FWRITE_STATUS (status, 1);

		if (v->e_genome)
		{
			status = gdl_expr_genome_fwrite (stream, v->e_genome);
			GDL_FWRITE_STATUS (status, GDL_SUCCESS);
		}

		has = (v->sample_info) ? 'y' : 'n';
		status = fwrite (&has, sizeof(unsigned char), 1, stream);
		GDL_FWRITE_STATUS (status, 1);

		if (v->sample_info)
		{
			status = gdl_table_fwrite (stream, v->sample_info);
			GDL_FWRITE_STATUS (status, GDL_SUCCESS);
		}

		if (v->npop)
		{
			status = fwrite (&(v->npop), sizeof(size_t), 1, stream);
			GDL_FWRITE_STATUS (status, 1);
			for(i = 0; i < v->npop; i++)
			{
				status = gdl_string_fwrite (stream, v->pop_names[i]);
				GDL_FWRITE_STATUS (status, GDL_SUCCESS);
			}
		}

		return GDL_SUCCESS;
	}

	return GDL_EINVAL;
}

FILE *
gdl_igdb_genome_set_logger (gdl_igdb_genome * g, FILE * stream)
{
	FILE * old = g->logger;
	g->logger = stream;
	return old;
}

int
gdl_igdb_genome_load_feature (gdl_igdb_genome * g, const gdl_string * feature_file)
{
	FILE * stream;

	stream = gdl_fileopen (feature_file, "r");
	g->f_genome = gdl_expr_feature_genome_fread (stream);
	gdl_fileclose (feature_file, stream);

	return (g->f_genome) ? GDL_SUCCESS : GDL_FAILURE;
}

int
gdl_igdb_genome_load_cis_feature (gdl_igdb_genome * g, const gdl_string * cis_feature_file)
{
	FILE * stream;

	stream = gdl_fileopen (cis_feature_file, "r");
	g->r_genome = gdl_expr_feature_cis_genome_fread (stream);
	gdl_fileclose (cis_feature_file, stream);

	return GDL_SUCCESS;
}

int
gdl_igdb_genome_summary (FILE * stream, const gdl_igdb_genome * v)
{
	fprintf (stdout, "<data-base>\n");
	fprintf (stdout, "   <snp-data>\n");
	if (v->s_genome)
	{
		//gdl_expr_genome_summarize (stdout, v->s_genome);
	}
	fprintf (stdout, "   </snp-data>\n");
	fprintf (stdout, "   <exp-data>\n");
	if (v->e_genome)
	{
		gdl_expr_genome_summarize (stream, v->e_genome);
	}
	fprintf (stdout, "   </exp-data>\n");
	fprintf (stdout, "</data-base>\n");
}

gdl_igdb_genome *
gdl_igdb_genome_subset (gdl_igdb_genome * g, gdl_string ** sample_names, gdl_string ** sub_sample_names, const size_t nsub_sample, const gdl_string * output)
{
	size_t i;
	gdl_string * file;
	FILE * stream;
	gdl_igdb_genome * sub_g;

	sub_g = GDL_CALLOC (gdl_igdb_genome, 1);

	sub_g->nchrom = g->nchrom;
	sub_g->chroms = GDL_MALLOC (gdl_string *, g->nchrom);
	for(i = 0; i < g->nchrom; i++)
	{
		sub_g->chroms[i] = gdl_string_clone (g->chroms[i]);
	}
	sub_g->dbdir = gdl_string_clone (output);
	sub_g->npop  = g->npop;
	if (g->pop_names)
	{
		sub_g->pop_names = GDL_MALLOC (gdl_string *, g->npop);
		for(i = 0; i < g->npop; i++)
		{
			sub_g->pop_names[i] = gdl_string_clone (g->pop_names[i]);
		}
	}

	if (g->s_genome)
	{
		sub_g->s_genome = gdl_snp_genome_subset (g->s_genome, sample_names, sub_sample_names, nsub_sample, output);
		file   = gdl_string_sprintf ("%s_snp.db", output);
		stream = gdl_fileopen (file, "w");
		gdl_snp_genome_fwrite (stream, sub_g->s_genome);
		gdl_fileclose (file, stream);
		gdl_string_free (file);
	}
	if (g->e_genome)
	{
		sub_g->e_genome = gdl_expr_genome_subset (g->e_genome, sample_names, sub_sample_names, nsub_sample, output);
		file   = gdl_string_sprintf ("%s_exp.db", output);
		stream = gdl_fileopen (file, "w");
		gdl_expr_genome_fwrite (stream, sub_g->e_genome);
		gdl_fileclose (file, stream);
		gdl_string_free (file);
	}
	if (g->f_genome)
	{
		sub_g->f_genome = gdl_expr_feature_genome_subset (g->e_genome, g->f_genome, sample_names, sub_sample_names, nsub_sample, output);
		file   = gdl_string_sprintf ("%s.fdb", output);
		stream = gdl_fileopen (file, "w");
		gdl_expr_feature_genome_fwrite (stream, sub_g->f_genome);
		gdl_fileclose (file, stream);
		gdl_string_free (file);
	}
	if (g->sample_info)
	{
		sub_g->sample_info = gdl_table_row_subset (g->sample_info, sub_sample_names, nsub_sample);
	}

	return sub_g;
}

int
gdl_igdb_genome_fprintf (gdl_igdb_genome * g, const gdl_string * output)
{
	if (g->s_genome)
	{
		gdl_snp_genome_fprintf (g->s_genome, g->pop_names, output);
	}
	if (g->e_genome)
	{
		gdl_expr_genome_fprintf (g->e_genome, g->pop_names, output);
	}
	if (g->f_genome)
	{
		gdl_expr_feature_genome_fprintf (g->e_genome, g->f_genome, g->pop_names, output);
	}
	if (g->sample_info)
	{
		gdl_string * file = gdl_string_sprintf ("%s/SampleInfo.txt", output);
		FILE * stream = gdl_fileopen(file, "w");
		gdl_table_fprintf (stream, g->sample_info);
		gdl_fileclose(file, stream);
		gdl_string_free (file);
	}
}
